#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2026 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import libadcc
import numpy as np

from .GroundState import GroundState
from .MoSpaces import split_spaces
from .functions import direct_sum, einsum
from .misc import cached_member_function
from . import block as b


class LazyMp(GroundState):
    @cached_member_function()
    def td1(self, space: str) -> libadcc.Tensor:
        """T2 amplitudes"""
        hf = self.reference_state
        sp = split_spaces(space)
        assert all(s == b.v for s in sp[2:])
        eia = self.df(sp[0] + b.v)
        ejb = self.df(sp[1] + b.v)
        return (
            hf.eri(space) / direct_sum("ia+jb->ijab", eia, ejb).symmetrise((2, 3))
        )

    def ts2(self, space: str, apply_cvs: bool = False) -> libadcc.Tensor:
        """
        Computes the second-order MP singles amplitudes.
        """
        split = split_spaces(space)
        assert len(split) == 2 and split[1] == b.v
        if split[0] == b.o:
            return self.ts2_ov(apply_cvs=apply_cvs)
        elif split[0] == b.c:
            return self.ts2_cv(apply_cvs=apply_cvs)
        raise NotImplementedError(
            "Second-order MP singles amplitudes not implemented for space "
            f"'{space}'."
        )

    @cached_member_function()
    def ts2_ov(self, apply_cvs: bool = False) -> libadcc.Tensor:
        """
        Computes the ov block of the second-order MP singles amplitudes.
        """
        if apply_cvs and not self.has_core_occupied_space:
            raise RuntimeError("Cannot apply the CVS approximation to a "
                               "ground state build on top of a HF reference state "
                               "without a core space.")

        hf = self.reference_state
        t2oo = self.td1(b.oovv)
        denom = -self.df(b.ov)
        res = (
            # N^5: O^2V^3 / N^4: O^1V^3
            + 0.5 * einsum('jabc,ijbc->ia', hf.ovvv, t2oo)
            # N^5: O^3V^2 / N^4: O^2V^2
            + 0.5 * einsum('jkib,jkab->ia', hf.ooov, t2oo)
        )
        # additional terms since we don't apply the CVS approximation
        # for the ground state
        # (all of the following terms vanish within the approximation)
        if self.has_core_occupied_space and not apply_cvs:
            t2oc = self.td1(b.ocvv)
            t2cc = self.td1(b.ccvv)
            res += (
                # N^5: O^2V^2C^1 / N^4: O^1V^2C^1
                + 1.0 * einsum("jKab,jKib->ia", t2oc, hf.ocov)
                # N^5: O^1V^2C^2 / N^4: V^2C^2
                + 0.5 * einsum("JKab,ibJK->ia", t2cc, hf.ovcc)
                # N^5: O^1V^3C^1 / N^4: V^3C^1
                + 0.5 * einsum("iJbc,Jabc->ia", t2oc, hf.cvvv)
            )
        return (res / denom).evaluate()

    @cached_member_function()
    def ts2_cv(self, apply_cvs: bool = False) -> libadcc.Tensor:
        """
        Computes the cv block of the second-order MP singles amplitudes.
        """
        assert self.has_core_occupied_space
        if apply_cvs:
            raise RuntimeError("The 'cv' block of the second-order MP singles "
                               "amplitude vanishes within the CVS approximation."
                               "Probably you don't want to compute it at all.")
        hf = self.reference_state
        t2oo = self.td1(b.oovv)
        t2oc = self.td1(b.ocvv)
        t2cc = self.td1(b.ccvv)
        denom = -self.df(b.cv)
        res = (
            # N^5: O^1V^2C^2 / N^4: O^1V^2C^1
            + 1.0 * einsum("jKab,jKIb->Ia", t2oc, hf.occv)
            # N^5: V^2C^3 / N^4: V^2C^2
            + 0.5 * einsum("JKab,JKIb->Ia", t2cc, hf.cccv)
            # N^5: O^2V^2C^1 / N^4: O^2V^2
            + 0.5 * einsum("jkab,jkIb->Ia", t2oo, hf.oocv)
            # N^5: V^3C^2 / N^4: V^3C^1
            + 0.5 * einsum("IJbc,Jabc->Ia", t2cc, hf.cvvv)
            # N^5: O^1V^3C^1 / N^4: O^1V^3V
            - 0.5 * einsum("jIbc,jabc->Ia", t2oc, hf.ovvv)
        )
        return (res / denom).evaluate()

    @cached_member_function()
    def td2(self, space: str) -> libadcc.Tensor:
        """Return the T^D_2 term"""
        if space != b.oovv:
            raise NotImplementedError("T^D_2 term not implemented "
                                      f"for space {space}.")
        t2erit = self.t2eri(b.oovv, b.ov).transpose((1, 0, 2, 3))
        denom = direct_sum(
            'ia,jb->ijab', self.df(b.ov), self.df(b.ov)
        ).symmetrise(0, 1)
        return (
            + 4.0 * t2erit.antisymmetrise(2, 3).antisymmetrise(0, 1)
            - 0.5 * self.t2eri(b.oovv, b.vv)
            - 0.5 * self.t2eri(b.oovv, b.oo)
        ) / denom

    @cached_member_function()
    def tt2(self, space: str) -> libadcc.Tensor:
        """
        Return the second order MP triples amplitudes for the given space
        (e.g. o1o1o1v1v1v1).
        """
        if space != b.ooovvv:
            raise NotImplementedError("Second order MP triples amplitudes not "
                                      f"implemented for space {space}.")
        hf = self.reference_state
        df = self.df(b.ov)
        t2_1 = self.t2(b.oovv)
        # denom = a + b + c - i - j - k   //   df = i - a
        denom = - 1 * direct_sum(
            "ia,jkbc->ijkabc", df, direct_sum("jb,kc->jkbc", df, df)
        ).symmetrise(0, 1, 2).symmetrise(3, 4, 5)
        # prefactor of 9, because we have 9 terms each in the expression, while the
        # antisymmetrisation generates 36 terms and introduces a prefactor of 1/36.
        # Each of the 9 terms is therefore generated 4 times.
        # The scaling in the comments is given as: [comp_scaling] / [mem_scaling]
        numerator = (
            # N^7: O^3V^4 / N^6: O^3V^3
            + 9 * einsum('idab,jkcd->ijkabc', hf.ovvv, t2_1)
            # N^7: O^4V^3 / N^6: O^3V^3
            + 9 * einsum('ijla,klbc->ijkabc', hf.ooov, t2_1)
        ).antisymmetrise(0, 1, 2).antisymmetrise(3, 4, 5)
        return numerator / denom

    @cached_member_function()
    def energy_correction(self, level: int = 2) -> float:
        """Obtain the MP energy correction at a particular level"""
        assert level >= 0
        if level < 2:
            return 0.0
        hf = self.reference_state
        is_cvs = self.has_core_occupied_space
        if level == 2 and not is_cvs:
            terms = [(1.0, hf.oovv, self.t2oo)]
        elif level == 2 and is_cvs:
            terms = [(1.0, hf.oovv, self.t2oo),
                     (2.0, hf.ocvv, self.t2oc),
                     (1.0, hf.ccvv, self.t2cc)]
        elif level == 3 and not is_cvs:
            terms = [(1.0, hf.oovv, self.td2(b.oovv))]
        else:
            method = "CVS-MP" if is_cvs else "MP"
            raise NotImplementedError(f"{method} energy correction for level "
                                      f"{level} not implemented.")
        return sum(
            -0.25 * pref * eri.dot(t2)
            for pref, eri, t2 in terms
        )

    def energy(self, level: int = 2) -> float:
        """
        Obtain the total energy (SCF energy plus all corrections)
        at a particular level of perturbation theory.
        """
        assert level >= 0
        if level == 0:
            # Sum of orbital energies ...
            return np.einsum(
                "i->", self.reference_state.foo.diagonal().to_ndarray()
            )

        # Accumulator for all energy terms
        energies = [self.reference_state.energy_scf]
        for il in range(2, level + 1):
            energies.append(self.energy_correction(il))
        return sum(energies)

    def to_qcvars(self, properties: bool = False,
                  recurse: bool = False, maxlevel: int = 2) -> dict:
        """
        Return a dictionary with property keys compatible to a Psi4 wavefunction
        or a QCEngine Atomicresults object.
        """
        return self._to_qcvars(
            gs_type="MP", properties=properties, recurse=recurse,
            maxlevel=maxlevel
        )

    @property
    def mp2_diffdm(self):
        """
        Return the MP2 difference density in the MO basis.
        """
        return self.diffdm(2)

    @property
    def mp2_density(self):
        return self.density(2)

    @property
    def mp2_dipole_moment(self):
        return self.dipole_moment(level=2)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import sys
import time
import numpy as np

from contextlib import contextmanager
from os.path import join


def strtime_short(span):
    """
    Return a 5-character string identifying the timespan
    """
    if span < 1:
        return "{:3d}ms".format(int(span * 1000))
    if span < 60:
        return "{:4.1f}s".format(span)
    if span < 120:
        return "{:4d}s".format(int(span))
    if span < 3600:
        return "{:4.1f}m".format(span / 60)
    if span < 86400:
        return "{:4.1f}h".format(span / 3600)
    else:
        return "{:4.1f}d".format(span / 3600 / 24)


def strtime(span, colour=False):
    """
    Return a moderately long string, providing a human-interpretable
    representation of the provided timespan
    """
    if colour:
        cms = "\033[0m"
        cs = "\033[38;5;246m"
        cm = "\033[38;5;239m"
        ch = "\033[38;5;250m"
        cd = "\033[38;5;242m"
        cr = "\033[0m"
    else:
        cr = cd = ch = cm = cs = cms = ""

    if span < 1:
        return cms + "{:6.3f}ms".format(span * 1000) + cr
    if span < 120:
        full = int(span)
        return (cs + "{:3d}s ".format(full) + cms
                + "{:3d}ms".format(int((span - full) * 1000)) + cr)
    if span < 3600:
        full = int(span / 60)
        return (cm + "{:2d}m ".format(full) + cs
                + "{:2d}s".format(int(span - full * 60)) + cr)
    if span < 86400:
        full = int(span / 3600)
        return (ch + "{:2d}h ".format(full) + cm
                + "{:2d}m".format(int(span / 60 - full * 60)) + cr)
    else:
        full = int(span / 86400)
        return (cd + "{:3d}d ".format(full) + ch
                + "{:2d}h".format(int(span / 3600 - full * 24)) + cr)


class Timer:
    # TODO More flexible: Time on a subtree
    # TODO Describe function to print a nice table
    def __init__(self):
        self.time_construction = time.perf_counter()
        self.raw_data = {}  # Raw data of time intervals [start, end]
        self.start_times = {}

    def attach(self, other, subtree=""):
        """
        Attach the timing results from another timer,
        i.e. merge both timers together.
        """
        for k, v in other.raw_data.items():
            kfull = join(subtree, k)
            if kfull not in self.raw_data:
                self.raw_data[kfull] = []
            self.raw_data[kfull].extend(v)

        for k, v in other.start_times.items():
            kfull = join(subtree, k)
            if kfull not in self.start_times:
                self.start_times[kfull] = []
            self.start_times[kfull].extend(v)

    def stop(self, task, now=None):
        """Stop a task and return runtime of it."""
        if now is None:
            now = time.perf_counter()
        if task in self.start_times:
            start = self.start_times.pop(task)
            if task in self.raw_data:
                self.raw_data[task].append((start, now))
            else:
                self.raw_data[task] = [(start, now)]
            return now - start
        return 0

    def restart(self, task):
        """
        Start a task if it is currently not running
        or stop and restart otherwise.
        """
        now = time.perf_counter()
        if self.is_running(task):
            self.stop(task, now)
        self.start_times[task] = now

    @contextmanager
    def record(self, task):
        """
        Context manager to automatically start and stop a time
        recording as long as context is active.

        Parameters
        ----------
        task : str
            The string describing the task
        """
        self.restart(task)
        try:
            yield self
        finally:
            self.stop(task)

    def is_running(self, task):
        return task in self.start_times

    @property
    def tasks(self):
        """The list of all tasks known to this object"""
        all_tasks = set(self.start_times.keys())
        all_tasks.update(set(self.raw_data.keys()))
        return sorted(list(all_tasks))

    @property
    def lifetime(self):
        """Get total time since this class has been constructed"""
        return time.perf_counter() - self.time_construction

    def intervals(self, task):
        """Get all time intervals recorded for a particular task"""
        if task not in self.raw_data:
            if task not in self.start_times:
                raise ValueError("Unknown task: " + task)
            return self.current(task)
        else:
            intervals = [end - start for start, end in self.raw_data[task]]
            if task in self.start_times:
                intervals.append(time.perf_counter() - self.start_times[task])
            return np.array(intervals)

    def total(self, task):
        """Get total runtime on a task in seconds"""
        return np.sum(self.intervals(task))

    def best(self, task):
        """Get the best time of all the intervals recorded for a task"""
        return np.min(self.intervals(task))

    def median(self, task):
        return np.median(self.intervals(task))

    def average(self, task):
        return np.average(self.intervals(task))

    def current(self, task):
        """Get current time on a task without stopping it"""
        if self.is_running(task):
            return time.perf_counter() - self.start_times[task]
        else:
            raise ValueError("Task not currently running: " + task)

    def describe(self, colour=sys.stdout.isatty()) -> str:
        # This is very dummy ... in the future we would like to have
        # a nice little table, which also respects the key hierachy
        # and displays cumulative sums for each level and which
        # tasks care of duplicated time intervals (e.g. if two tasks
        # run at the same time)
        # Colour: Use one for each level
        text = "Timer " + strtime_short(self.lifetime) + " lifetime:\n"
        if not self.tasks:
            text += "  No timings recorded"
            return text

        maxlen = max(len(key) for key in self.tasks)
        for key in self.tasks:
            fmt = "  {:<" + str(maxlen) + "s} {:>20s}\n"
            text += fmt.format(key, strtime(self.total(key), colour=colour))
        return text

    def _repr_pretty_(self, pp, cycle):
        if cycle:
            pp.text("Timer()")
        else:
            pp.text(self.describe())


Timer.start = Timer.restart


def timed_call(f):
    """
    Decorator to automatically time function calls.
    The timer object is available under the function attribute _timer
    """
    def decorated(*args, **kwargs):
        if not hasattr(decorated, "_timer"):
            setattr(decorated, "_timer", Timer())
        with getattr(decorated, "_timer").record(f.__name__):
            return f(*args, **kwargs)
    decorated.__doc__ = f.__doc__
    return decorated


def timed_member_call(timer="timer"):
    """
    Decorator to automatically time calls to instance member functions.
    The name of the instance attribute where timings are stored is the
    timer argument to this function.
    """
    def decorator(f):
        def wrapped(self, *args, **kwargs):
            if not hasattr(self, timer):
                setattr(self, timer, Timer())
            with getattr(self, timer).record(f.__name__):
                return f(self, *args, **kwargs)
        wrapped.__doc__ = f.__doc__
        return wrapped
    return decorator



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc
import psi4

from matplotlib import pyplot as plt

# Run SCF in psi4
mol = psi4.geometry("""
    O 0 0 0
    H 0 0 1.795239827225189
    H 1.693194615993441 0 -0.599043184453037
    symmetry c1
    units au
    no_reorient
    no_com
    """)

# set the number of cores equal to the auto-determined value from
# the adcc ThreadPool
psi4.set_num_threads(adcc.get_n_threads())
psi4.core.be_quiet()
psi4.set_options({'basis': "cc-pvdz",
                  'scf_type': 'pk',
                  'e_convergence': 1e-14,
                  'd_convergence': 1e-9})
scf_e, wfn = psi4.energy('SCF', return_wfn=True)

print(adcc.banner())

# Run an adc2 calculation:
state = adcc.adc2(wfn, n_singlets=7, conv_tol=1e-8)

# Print results
print()
print("  st  ex.ene. (au)         f     transition dipole moment (au)"
      "        state dip (au)")
for i, val in enumerate(state.excitation_energy):
    fmt = "{0:2d}  {1:12.8g} {2:9.3g}   [{3:9.3g}, {4:9.3g}, {5:9.3g}]"
    fmt += "   [{6:9.3g}, {7:9.3g}, {8:9.3g}]"
    print(state.kind[0], fmt.format(i, val, state.oscillator_strength[i],
                                    *state.transition_dipole_moment[i],
                                    *state.state_dipole_moment[i]))

state.plot_spectrum()
plt.savefig("psi4_ccpvdz_adc2_spectrum.pdf")
plt.show()

# Print timings summary:
print()
print(state.timer.describe())



#!/usr/bin/env python3
import sys
import numpy as np
import scipy.linalg as la

from adcc import copy, evaluate

from ..functions import dot
from .preconditioner import PreconditionerIdentity
from .explicit_symmetrisation import IndexSymmetrisation


def guess_from_previous(matrix, rhs, previous_cgstate):
    """
    Gets passed the matrix, the cgstate from the previous solution
    and the new RHS and may use this information to construct
    a guess for the next CG to perform.
    """
    if previous_cgstate is None:
        return rhs
    else:
        return previous_cgstate.solution


def guess_from_rhs(matrix, rhs, previous_cgstate):
    return rhs


class IterativeInverse:
    def __init__(self, matrix, construct_guess=guess_from_rhs, **kwargs):
        """Initialise an iterative inverse

        This object mimics to be the inverse of a passed matrix
        by solving linear system's of equations each time the `@`
        operator is used in conjunction with this matrix.

        Parameters
        ----------
        matrix
            Matrix object
        """
        self.matrix = matrix
        self.kwargs = kwargs
        self.construct_guess = construct_guess
        self.cgstate = None

    @property
    def shape(self):
        return self.matrix.shape  # Inversion does not change the shape

    def __matmul__(self, x):
        if isinstance(x, list):
            return [self.__matmul__(xi) for xi in x]
        else:
            guess = self.construct_guess(self.matrix,  x, self.cgstate)
            self.cgstate = conjugate_gradient(self.matrix, x, guess,
                                              **self.kwargs)
            return self.cgstate.solution


class State:
    def __init__(self):
        self.solution = None       # Current approximation to the solution
        self.residual = None       # Current residual
        self.residual_norm = None  # Current residual norm
        self.converged = False     # Flag whether iteration is converged
        self.n_iter = 0            # Number of iterations
        self.n_applies = 0         # Number of applies


def default_print(state, identifier, file=sys.stdout):
    if identifier == "start" and state.n_iter == 0:
        print("Niter residual_norm", file=file)
    elif identifier == "next_iter":
        fmt = "{n_iter:3d}  {residual:12.5g}"
        print(fmt.format(n_iter=state.n_iter,
                         residual=np.max(state.residual_norm)), file=file)
    elif identifier == "is_converged":
        print("=== Converged ===", file=file)
        print("    Number of matrix applies:   ", state.n_applies)


def conjugate_gradient(matrix, rhs, x0=None, conv_tol=1e-9, max_iter=100,
                       callback=None, Pinv=None, cg_type="polak_ribiere",
                       explicit_symmetrisation=IndexSymmetrisation):
    """An implementation of the conjugate gradient algorithm.

    This algorithm implements the "flexible" conjugate gradient using the
    Polak-Ribière formula, but allows to employ the "traditional"
    Fletcher-Reeves formula as well.
    It solves `matrix @ x = rhs` for `x` by minimising the residual
    `matrix @ x - rhs`.

    Parameters
    ----------
    matrix
        Matrix object. Should be an ADC matrix.
    rhs
        Right-hand side, source.
    x0
        Initial guess
    conv_tol : float
        Convergence tolerance on the l2 norm of residuals to consider
        them converged.
    max_iter : int
        Maximum number of iterations
    callback
        Callback to call after each iteration
    Pinv
        Preconditioner to A, typically an estimate for A^{-1}
    cg_type : string
        Identifier to select between polak_ribiere and fletcher_reeves
    explicit_symmetrisation
        Explicit symmetrisation to perform during iteration to ensure
        obtaining an eigenvector with matching symmetry criteria.
    """
    if callback is None:
        def callback(state, identifier):
            pass

    if explicit_symmetrisation is not None and \
            isinstance(explicit_symmetrisation, type):
        explicit_symmetrisation = explicit_symmetrisation(matrix)

    if x0 is None:
        # Start with random guess
        raise NotImplementedError("Random guess is not yet implemented.")
    else:
        x0 = copy(x0)

    if Pinv is None:
        Pinv = PreconditionerIdentity()
    if Pinv is not None and isinstance(Pinv, type):
        Pinv = Pinv(matrix)

    def is_converged(state):
        state.converged = state.residual_norm < conv_tol
        return state.converged

    state = State()

    # Initialise iterates
    state.solution = x0
    state.residual = evaluate(rhs - matrix @ state.solution)
    state.n_applies += 1
    state.residual_norm = np.sqrt(state.residual @ state.residual)
    pk = zk = Pinv @ state.residual

    if explicit_symmetrisation:
        # TODO Not sure this is the right spot ... also this syntax is ugly
        pk = explicit_symmetrisation.symmetrise(pk)

    callback(state, "start")
    while state.n_iter < max_iter:
        state.n_iter += 1

        # Update ak and iterated solution
        # TODO This needs to be modified for general optimisations,
        #      i.e. where A is non-linear
        # https://en.wikipedia.org/wiki/Nonlinear_conjugate_gradient_method
        Apk = matrix @ pk
        state.n_applies += 1
        res_dot_zk = dot(state.residual, zk)
        ak = float(res_dot_zk / dot(pk, Apk))
        state.solution = evaluate(state.solution + ak * pk)

        residual_old = state.residual
        state.residual = evaluate(residual_old - ak * Apk)
        state.residual_norm = np.sqrt(state.residual @ state.residual)

        callback(state, "next_iter")
        if is_converged(state):
            state.converged = True
            callback(state, "is_converged")
            return state

        if state.n_iter == max_iter:
            raise la.LinAlgError("Maximum number of iterations (== "
                                 + str(max_iter) + " reached in conjugate "
                                 "gradient procedure.")

        zk = evaluate(Pinv @ state.residual)

        if explicit_symmetrisation:
            # TODO Not sure this is the right spot ... also this syntax is ugly
            zk = explicit_symmetrisation.symmetrise(zk)

        if cg_type == "fletcher_reeves":
            bk = float(dot(zk, state.residual) / res_dot_zk)
        elif cg_type == "polak_ribiere":
            bk = float(dot(zk, (state.residual - residual_old)) / res_dot_zk)
        pk = zk + bk * pk



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import os
import h5py
import warnings

from ..misc import is_module_available

__all__ = ["import_scf_results", "run_hf", "have_backend", "available"]


# Lazily cache the available backends
_available_backends = None


def available() -> tuple[str, ...]:
    global _available_backends

    if _available_backends is None:
        status = {
            "pyscf": is_module_available("pyscf", "1.5.0"),
            "psi4": (is_module_available("psi4", "1.3.0")
                     and is_module_available("psi4.core")),
            "veloxchem": is_module_available("veloxchem"),  # No version info
            "molsturm": is_module_available("molsturm"),    # No version info
        }
        _available_backends = tuple(sorted(
            b for b, stat in status.items() if stat
        ))
    return _available_backends


def have_backend(backend):
    """Is a particular backend available?"""
    return backend in available()


def import_scf_results(res):
    """
    Import an scf_result from an SCF program. Tries to be smart
    and guess what the host program was and how to import it.
    """
    if have_backend("pyscf"):
        from . import pyscf as backend_pyscf
        from pyscf import scf

        if isinstance(res, scf.hf.SCF):
            return backend_pyscf.import_scf(res)

    if have_backend("molsturm"):
        from . import molsturm as backend_molsturm
        from molsturm.State import State

        if isinstance(res, State):
            return backend_molsturm.import_scf(res)

    if have_backend("veloxchem"):
        import veloxchem as vlx

        from . import veloxchem as backend_veloxchem

        if isinstance(res, vlx.scfrestdriver.ScfDriver):
            return backend_veloxchem.import_scf(res)

    if have_backend("psi4"):
        import psi4
        from . import psi4 as backend_psi4

        if isinstance(res, psi4.core.HF):
            return backend_psi4.import_scf(res)

    from libadcc import HartreeFockSolution_i
    if isinstance(res, HartreeFockSolution_i):
        return res

    if isinstance(res, (dict, h5py.File)):
        from adcc.DataHfProvider import DataHfProvider
        return DataHfProvider(res)

    if isinstance(res, str):
        if os.path.isfile(res) and (res.endswith(".h5")
                                    or res.endswith(".hdf5")):
            return import_scf_results(h5py.File(res, "r"))
        else:
            raise ValueError("Unrecognised path or file extension: {}"
                             "".format(res))

    # Note: Add more backends here

    raise NotImplementedError("No means to import an SCF result of "
                              "type " + str(type(res)) + " implemented.")


def run_hf(backend, xyz, basis, **kwargs):
    """
        Run a HF calculation with a specified backend, molecule, and SCF
        parameters

        backend:        name of the backend (pyscf, psi4, or veloxchem)
        xyz:            string with coordinates in Bohr
        basis:          basis set name
        charge:         charge of the molecule
        multiplicity:   spin multiplicity 2S + 1
        conv_tol:       energy convergence tolerance
        conv_tol_grad:  convergence tolerance of the electronic gradient
        max_iter:       maximum number of SCF iterations
        pe_options:     dictionary with options for polarizable embedding (PE)

        Note: This function only exists for testing purposes and should
        not be used in production calculations.
    """

    if not backend:
        if len(available()) == 0:
            raise RuntimeError(
                "No supported host-program available as SCF backend. "
                "See https://adc-connect.org/latest/"
                "installation.html#install-hostprogram "
                "for installation instructions."
            )
        else:
            backend = available()[0]
        warnings.warn("No backend specified. Using {}.".format(backend))

    if not have_backend(backend):
        raise ValueError("Backend {} not found.".format(backend))
    if backend == "psi4":
        from . import psi4 as backend_hf

    elif backend == "pyscf":
        from . import pyscf as backend_hf

    elif backend == "veloxchem":
        from . import veloxchem as backend_hf

    elif backend == "molsturm":
        from . import molsturm as backend_hf

    else:
        raise NotImplementedError("No run_hf function implemented for backend "
                                  "{}.".format(backend))

    return backend_hf.run_hf(xyz, basis, **kwargs)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

from pyscf import gto, scf
from matplotlib import pyplot as plt

# Run SCF in pyscf
mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='cc-pvdz',
    unit="Bohr"
)
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-12
scfres.conv_tol_grad = 1e-9
scfres.kernel()

print(adcc.banner())

# Run an adc2 calculation:
state = adcc.adc2(scfres, n_singlets=7, conv_tol=1e-8)

# Print results
print()
print("  st  ex.ene. (au)         f     transition dipole moment (au)"
      "        state dip (au)")
for i, val in enumerate(state.excitation_energy):
    fmt = "{0:2d}  {1:12.8g} {2:9.3g}   [{3:9.3g}, {4:9.3g}, {5:9.3g}]"
    fmt += "   [{6:9.3g}, {7:9.3g}, {8:9.3g}]"
    print(state.kind[0], fmt.format(i, val, state.oscillator_strength[i],
                                    *state.transition_dipole_moment[i],
                                    *state.state_dipole_moment[i]))
print(state.excitation_energies)
state.plot_spectrum()
plt.savefig("pyscf_ccpvdz_adc2_spectrum.pdf")
plt.show()

# Print timings summary:
print()
print(state.timer.describe())



#!/usr/bin/env python3

# Generate valid IF_DIMENSIONS_MATCH_EXECUTE_CONTRACT lines for TensorImpl.cc

# The maximal tensor dimensionality
maxdim = 6


def is_valid(n_contr_idcs, dima, dimb, dimout):
    return dima > 0 and dimb > 0 and \
        dima >= n_contr_idcs and dimb >= n_contr_idcs and dimout > 0 and \
        dimout + n_contr_idcs + n_contr_idcs == dima + dimb


valid_combinations = []
for n_contr_idcs in range(maxdim + 1):
    for dima in range(maxdim + 1):
        for dimb in range(maxdim + 1):
            for dimout in range(maxdim + 1):
                if is_valid(n_contr_idcs, dima, dimb, dimout):
                    valid_combinations.append(
                        (n_contr_idcs, dima, dimb, dimout)
                    )

print("//")
print("// Instantiation generated from TensorImpl/instantiate_valid.py")
print("//")
for n_contr_idcs, dima, dimb, dimout in valid_combinations:
    if n_contr_idcs == 0:
        continue
    print(f"IF_DIMENSIONS_MATCH_EXECUTE_CONTRACT({n_contr_idcs}, {dima}, {dimb}) //")  # noqa: E501

print()
print()
print("//")
print("// Instantiation generated from TensorImpl/instantiate_valid.py")
print("//")
for n_contr_idcs, dima, dimb, dimout in valid_combinations:
    if n_contr_idcs != 0:
        continue
    print(f"IF_DIMENSIONS_MATCH_EXECUTE_TENSORPROD({dima}, {dimb}) //")



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from .MoSpaces import MoSpaces
from .ReferenceState import ReferenceState


class FormatIndexBase:
    def __init__(self, *args, **kwargs):
        pass

    def optimise_formatting(self, *space_index_pairs: tuple[str, int]) -> None:
        pass

    def format(self, space, idx, concat_spin=True):
        raise NotImplementedError("Implement the format function.")

    @property
    def max_n_characters(self) -> int:
        raise NotImplementedError("Implement the max_n_character method.")


class FormatIndexHfProvider(FormatIndexBase):
    def __init__(self, refstate: ReferenceState, max_digits: int = 1):
        """
        Format the given index for the given orbital subspace
        looking up it's equivalent index in the HF Provider's
        indexing convention. Returns a tuple, namely the formatted
        index and the corresponding spin block.

        Parameters
        ----------
        refstate : ReferenceState
            ReferenceState to use to get host program information
        max_digits : int, optional
            Number of digits to reserve for the host index
        """
        if not isinstance(refstate, ReferenceState):
            raise TypeError(f"Unsupported type: {str(refstate)}")
        # TODO some backends such as pyscf have a 1-based indexing
        #      ... this should be worked in here!
        self.base_index = 0
        self.mospaces = refstate.mospaces
        self.max_digits = max_digits
        self.noa: int = self.mospaces.n_orbs_alpha("f")

    def _translate_index(self, space: str, idx: int) -> tuple[int, str]:
        ihf = self.mospaces.map_index_hf_provider[space][idx]
        if ihf < self.noa:
            return self.base_index + ihf, "a"
        else:
            return self.base_index + ihf - self.noa, "b"

    def optimise_formatting(self, *space_index_pairs: tuple[str, int]) -> None:
        """
        Optimise the formatting parameters of this class in order to be able to
        nicely produce equivalently formatted tensor format strings for all the
        passed spaces-index pairs.

        This function can be called multiple times.
        """
        log_max_idx = int(np.log(max(self._translate_index(space, idx)[0]
                                     for space, idx in space_index_pairs)))
        self.max_digits = max(log_max_idx, self.max_digits, 1)

    def format(self, space: str, idx: int,
               concat_spin: bool = True) -> tuple[str, str]:
        """
        Format the given index that belongs to the given space.
        Return a tuple containing the formatted index and the corresponding
        formatted spin.
        """
        fstr = "{:" + str(self.max_digits) + "d}"
        tidx, spin = self._translate_index(space, idx)
        if concat_spin:
            return fstr.format(tidx) + spin
        else:
            return fstr.format(tidx), spin

    @property
    def max_n_characters(self) -> int:
        """
        The maximum number of characters needed for a formatted index (excluding
        the "a" or "b" spin string) according to the current optimised formatting.
        """
        return self.max_digits


class FormatIndexAdcc(FormatIndexBase):
    def __init__(self, refstate_or_mospaces, max_digits: int = 1):
        """
        Format the given index for the given orbital subspace
        by simply pretty-printing the passed space and index data.

        Parameters
        ----------
        refstate_or_mospaces : ReferenceState or MoSpaces
            ReferenceState to use to get host program information
        max_digits : int, optional
            Number of digits to reserve for the host index
        """
        if isinstance(refstate_or_mospaces, ReferenceState):
            self.mospaces = refstate_or_mospaces.mospaces
        elif isinstance(refstate_or_mospaces, MoSpaces):
            self.mospaces = refstate_or_mospaces
        else:
            raise TypeError(f"Unsupported type: {str(refstate_or_mospaces)}")
        self.max_digits = max_digits

    def _translate_index(self, space: str, idx: int) -> tuple[str, int, str]:
        noa = self.mospaces.n_orbs_alpha(space)
        if idx < noa:
            return space, idx, "a"
        else:
            return space, idx - noa, "b"

    def optimise_formatting(self, *space_index_pairs: tuple[str, int]):
        """
        Optimise the formatting parameters of this class in order to be able to
        nicely produce equivalently formatted tensor format strings for all the
        passed spaces-index pairs.

        This function can be called multiple times.
        """
        maxlen = max(self._translate_index(space, idx)[1]
                     for space, idx in space_index_pairs)
        log_max_idx = int(np.log(max(1, maxlen)))
        self.max_digits = max(log_max_idx, self.max_digits, 1)

    def format(self, space: str, idx: int,
               concat_spin: bool = True) -> tuple[str, str]:
        """
        Format the given index that belongs to the given space.
        Return a tuple containing the formatted index and the corresponding
        formatted spin.
        """
        fstr = "({:2s} {:" + str(self.max_digits) + "d})"
        space, tidx, spin = self._translate_index(space, idx)
        if concat_spin:
            return fstr.format(space, tidx) + spin
        else:
            return fstr.format(space, tidx), spin

    @property
    def max_n_characters(self) -> int:
        """
        The maximum number of characters needed for a formatted index (excluding
        the "a" or "b" spin string) according to the current optimised formatting.
        """
        return 5 + self.max_digits


class FormatIndexHomoLumo(FormatIndexBase):
    def __init__(self, refstate: ReferenceState, max_digits: int = 1,
                 use_hoco: bool = True):
        """
        Format the given index for the given orbital subspace
        by translating it into a form HOMO-1, LUMO+4 and so on.
        The special label HOCO can be optionally used to refer
        to the highest occupied orbital of the core space
        of CVS calculations.

        Parameters
        ----------
        refstate : adcc.ReferenceState
            ReferenceState to use to get host program information
        max_digits : int, optional
            Number of digits to reserve for the HOMO/LUMO/HOCO
            offset.
        use_hoco : bool, optional
            Use the special HOCO indicator to point to the highest-energy
            orbital of the CVS core space (True) or treat it relative
            to the HOMO as well (False).
        """
        if not isinstance(refstate, ReferenceState):
            raise TypeError(f"Unsupported type: {str(refstate)}")
        self.mospaces: MoSpaces = refstate.mospaces
        self.maxlen_offset: int = max_digits + 1  # + 1 for "+"/"-" string
        self.n_alpha: int = refstate.n_alpha
        self.n_beta: int = refstate.n_beta
        self.use_hoco = use_hoco

        if not self.mospaces.has_core_occupied_space:
            self.use_hoco = False

        # I do not know the convention or whether this even makes
        # sense in the cases where there is no closed-shell or no Aufbau
        # TODO Expand this class there is something sensible
        closed_shell = refstate.n_alpha == refstate.n_beta
        if not refstate.is_aufbau_occupation or not closed_shell:
            raise ValueError("format_homolumo only produces the right results "
                             "for closed-shell references with an Aufbau "
                             "occupation")

    def _translate_index(self, space: str, idx: int) -> tuple[str, str, str]:
        # Deal with core-occupied orbitals first:
        if self.use_hoco and space == "o2":
            noa = self.mospaces.n_orbs_alpha("o2")
            no = self.mospaces.n_orbs("o2")
            if idx == noa - 1:
                return "HOCO", "", "a"
            elif idx == no - 1:
                return "HOCO", "", "b"
            elif idx < noa:
                return "HOCO", "-" + str(noa - 1 - idx), "a"
            else:
                return "HOCO", "-" + str(no - 1 - idx), "b"
        else:
            ihf = self.mospaces.map_index_hf_provider[space][idx]
            ifull = self.mospaces.map_index_hf_provider["f"].index(ihf)
            spin = "a"
            noa = self.mospaces.n_orbs_alpha("f")
            assert noa == self.mospaces.n_orbs_beta("f")
            if ifull >= noa:
                spin = "b"
                ifull = ifull - noa

            ihomo = self.n_alpha - 1
            ilumo = ihomo + 1
            if ifull < ihomo:
                return "HOMO", "-" + str(ihomo - ifull), spin
            elif ifull == ihomo:
                return "HOMO", "", spin
            elif ifull == ilumo:
                return "LUMO", "", spin
            elif ifull > ilumo:
                return "LUMO", "+" + str(ifull - ilumo), spin

    def optimise_formatting(self, *space_index_pairs: tuple[str, int]) -> None:
        """
        Optimise the formatting parameters of this class in order to be able to
        nicely produce equivalently formatted tensor format strings for all the
        passed spaces-index pairs.

        This function can be called multiple times.
        """
        maxlen = max(len(self._translate_index(space, idx)[1])
                     for space, idx in space_index_pairs)
        self.maxlen_offset = max(maxlen, self.maxlen_offset, 2)

    def format(self, space: str, idx: int,
               concat_spin: bool = True) -> tuple[str, str]:
        """
        Format the given index that belongs to the given space.
        Return a tuple containing the formatted index and the corresponding
        formatted spin.
        """
        word, offset, spin = self._translate_index(space, idx)
        fmt = "{:s}{:>" + str(self.maxlen_offset) + "s}"
        if concat_spin:
            return fmt.format(word, offset) + spin
        else:
            return fmt.format(word, offset), spin

    @property
    def max_n_characters(self) -> int:
        """
        The maximum number of characters needed for a formatted index (excluding
        the "a" or "b" spin string) according to the current optimised formatting.
        """
        return 4 + self.maxlen_offset  # +4 for HOMO/LUMO/HOCO



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import sys

from libadcc import HartreeFockProvider, get_n_threads, set_n_threads

from .LazyMp import LazyMp
from .Tensor import Tensor
from .Symmetry import Symmetry
from .MoSpaces import MoSpaces
from .AdcMatrix import AdcMatrix
from .AdcMethod import AdcMethod, IsrMethod
from .functions import (copy, direct_sum, dot, einsum, empty_like,
                        evaluate, lincomb, linear_combination, nosym_like,
                        ones_like, transpose, zeros_like)
from .memory_pool import memory_pool
from .State2States import State2States
from .ExcitedStates import ExcitedStates
from .Excitation import Excitation
from .ElectronicTransition import ElectronicTransition
from .DataHfProvider import DataHfProvider, DictHfProvider
from .ReferenceState import ReferenceState
from .AmplitudeVector import AmplitudeVector
from .NParticleOperator import OperatorSymmetry
from .OneParticleOperator import OneParticleOperator
from .OneParticleDensity import OneParticleDensity
from .TwoParticleOperator import TwoParticleOperator
from .TwoParticleDensity import TwoParticleDensity
from .opt_einsum_integration import register_with_opt_einsum

# This has to be the last set of import
from .guess import (guess_symmetries, guess_zero, guesses_any, guesses_singlet,
                    guesses_spin_flip, guesses_triplet)
from .workflow import run_adc
from .exceptions import InputError

__all__ = ["run_adc", "InputError", "AdcMatrix",
           "AdcMethod", "IsrMethod", "Symmetry", "ReferenceState", "MoSpaces",
           "einsum", "copy", "dot", "empty_like", "evaluate",
           "lincomb", "nosym_like", "ones_like", "transpose",
           "linear_combination", "zeros_like", "direct_sum",
           "memory_pool", "set_n_threads", "get_n_threads", "AmplitudeVector",
           "HartreeFockProvider", "ExcitedStates", "State2States",
           "Excitation", "ElectronicTransition", "Tensor", "DictHfProvider",
           "DataHfProvider", "OneParticleOperator", "OneParticleDensity",
           "TwoParticleOperator", "TwoParticleDensity", "OperatorSymmetry",
           "guesses_singlet", "guesses_triplet", "guesses_any",
           "guess_symmetries", "guesses_spin_flip", "guess_zero", "LazyMp",
           "adc0", "cis", "adc1", "adc2", "adc2x", "adc3",
           "cvs_adc0", "cvs_adc1", "cvs_adc2", "cvs_adc2x", "cvs_adc3",
           "banner"]

__version__ = "0.18.0"
__license__ = "GPL v3"
__url__ = "https://adc-connect.org"
__authors__ = ["Michael F. Herbst", "Maximilian Scheurer", "Jonas Leitner",
               "Antonia Papapostolou", "Friederike Schneider",
               "Adrian L. Dempwolff", "Adrian J. Müller"]
__email__ = "developers@adc-connect.org"
__contributors__ = []


def with_runadc_doc(func):
    func.__doc__ = run_adc.__doc__
    return func


@with_runadc_doc
def adc0(*args, **kwargs):
    return run_adc(*args, **kwargs, method="adc0")


@with_runadc_doc
def cis(*args, **kwargs):
    state = run_adc(*args, **kwargs, method="adc1")
    return ExcitedStates(state, property_method="isr0")


@with_runadc_doc
def adc1(*args, **kwargs):
    return run_adc(*args, **kwargs, method="adc1")


@with_runadc_doc
def adc2(*args, **kwargs):
    return run_adc(*args, **kwargs, method="adc2")


@with_runadc_doc
def adc2x(*args, **kwargs):
    return run_adc(*args, **kwargs, method="adc2x")


@with_runadc_doc
def adc3(*args, **kwargs):
    return run_adc(*args, **kwargs, method="adc3")


@with_runadc_doc
def cvs_adc0(*args, **kwargs):
    return run_adc(*args, **kwargs, method="cvs-adc0")


@with_runadc_doc
def cvs_adc1(*args, **kwargs):
    return run_adc(*args, **kwargs, method="cvs-adc1")


@with_runadc_doc
def cvs_adc2(*args, **kwargs):
    return run_adc(*args, **kwargs, method="cvs-adc2")


@with_runadc_doc
def cvs_adc2x(*args, **kwargs):
    return run_adc(*args, **kwargs, method="cvs-adc2x")


@with_runadc_doc
def cvs_adc3(*args, **kwargs):
    return run_adc(*args, **kwargs, method="cvs-adc3")


def banner(colour=sys.stdout.isatty()):
    """Return a nice banner describing adcc and its components

    The returned string contains version information, maintainer emails
    and references.

    Parameters
    ----------
    colour : bool
        Should colour be used in the print out
    """
    if colour:
        yellow = '\033[93m'
        green = '\033[92m'
        cyan = '\033[96m'
        white = '\033[0m'
    else:
        yellow = ''
        green = ''
        cyan = ''
        white = ''

    empty = "|" + 70 * " " + "|\n"
    string = "+" + 70 * "-" + "+\n"
    string += "|{0:^70s}|\n".format(
        "adcc:  Seamlessly connect your host program to ADC"
    ).replace("adcc", "adc" + yellow + "c" + white)
    string += "+" + 70 * "-" + "+\n"
    string += empty
    string += "|     version     " + green + f"{__version__:<52}" + white + " |\n"

    # Print authors as groups
    groups = []
    cbuffer = []
    for i, author in enumerate(__authors__):
        if len(", ".join(cbuffer) + author) + 2 <= 52:
            cbuffer.append(author)
        else:
            groups.append(cbuffer)
            cbuffer = [author]
    if cbuffer:
        groups.append(cbuffer)
    for i, buf in enumerate(groups):
        authors = "authors" if i == 0 else ""
        joined = ", ".join(buf)
        if i != len(groups) - 1:
            joined += ","
        string += f"|     {authors:8s}    {joined:<52} |\n"

    string += "|     citation    " + yellow + "DOI 10.1002/wcms.1462" + white
    string += 32 * " " + "|\n"
    string += f"|     website     {__url__:<52} |\n"
    string += "|     email       " + cyan + f"{__email__:<52}" + white + " |\n"
    string += empty
    string += "+" + 70 * "-" + "+"
    return string


register_with_opt_einsum()



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

from pyscf import gto, scf

# Run SCF in pyscf
mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='sto-3g',
    unit="Bohr"
)

# Run RHF SCF
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-14
scfres.conv_tol_grad = 1e-10
scfres.kernel()

# Run an adc1 calculation:
state = adcc.adc1(scfres, n_singlets=5)

print(state.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------

"""Setup for adcc"""
import os
import sys
import json
import time
import shlex
import shutil
import logging
import platform
import tempfile
import functools
import sysconfig
import subprocess

from pathlib import Path

from setuptools import Command, setup

from pybind11.setup_helpers import Pybind11Extension
from pybind11.setup_helpers import build_ext as _Pybind11BuildExt


log = logging.getLogger()


#
# Custom commands
#
class BuildDocs(Command):
    description = "Build the C++ and python documentation"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        build_folder = Path("build")
        docs_folder = Path("docs")
        assert docs_folder.is_dir()

        try:  # generate the documentation for libadcc
            # we need to create the folder for doxygen
            (build_folder / "libadcc_docs").mkdir(parents=True, exist_ok=True)
            doxyfile = docs_folder / "Doxyfile"
            assert doxyfile.is_file()  # config file
            subprocess.check_call(["doxygen", str(doxyfile)])
        except (OSError, subprocess.CalledProcessError) as e:
            raise RuntimeError(
                f"Could not build C++ documentation with doxygen: {e}"
            )
        try:  # generate full documentation
            assert (docs_folder / "conf.py").is_file()  # config file
            output = build_folder / "docs"
            subprocess.check_call([
                "sphinx-build", "-M", "html", str(docs_folder), str(output)
            ])
        except (OSError, subprocess.CalledProcessError) as e:
            raise RuntimeError(
                f"Could not build adcc documentation with sphinx: {e}"
            )


class CppTest(Command):
    description = "Build and run C++ tests"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        test_executable = self.compile_test_executable()
        subprocess.check_call([test_executable])

    def compile_test_executable(self):
        from setuptools._distutils.ccompiler import new_compiler
        from setuptools._distutils.sysconfig import customize_compiler

        output_dir = Path("build") / "cpptest"
        test_executable = output_dir / "libadcc_tests"
        if test_executable.is_file():
            print(f"Skipping recompilation of {test_executable}. "
                  "Delete file if recompilation desired.")
            return test_executable  # Don't recompile

        # Download catch
        catch_header = output_dir / "catch2" / "catch.hpp"
        if not catch_header.is_file():
            (output_dir / "catch2").mkdir(parents=True, exist_ok=True)
            base = "https://github.com/catchorg/Catch2/releases/download/"
            request_urllib(base + "v2.13.9/catch.hpp", str(catch_header))

        # Adapt stuff from libadcc extension
        libadcc = libadcc_extension()
        include_dirs = libadcc.include_dirs + [str(output_dir)]
        sources = libadcc_sources("cpptest")
        extra_compile_args = [
            arg for arg in libadcc.extra_compile_args
            if not any(arg.startswith(start) for
                       start in ("-fvisibility", "-g", "-O"))
        ]

        # Reduce optimisation a bit to ensure that the debugging experience is good
        if "--coverage" in extra_compile_args:
            extra_compile_args += ["-O0", "-g"]
        else:
            extra_compile_args += ["-O1", "-g"]

        # Bring forward stuff from libadcc
        compiler = new_compiler()
        customize_compiler(compiler)
        objects = compiler.compile(
            sources, str(output_dir), libadcc.define_macros, include_dirs,
            debug=True, extra_postargs=extra_compile_args
        )

        if libadcc.extra_objects:
            objects.extend(libadcc.extra_objects)

        compiler.link_executable(
            objects, "libadcc_tests", str(output_dir), libadcc.libraries,
            libadcc.library_dirs, libadcc.runtime_library_dirs, debug=True,
            extra_postargs=libadcc.extra_link_args, target_lang=libadcc.language
        )
        return test_executable


#
# Setup libadcc extension
#
def append_to_pkg_config_path(*paths: str):
    """Append the given paths to the `PKG_CONFIG_PATH`."""
    for path in paths:
        if "PKG_CONFIG_PATH" in os.environ:
            os.environ["PKG_CONFIG_PATH"] += os.pathsep + path
        else:
            os.environ["PKG_CONFIG_PATH"] = path


@functools.lru_cache()
def get_pkg_config():
    """
    Get path to pkg-config and set up the PKG_CONFIG environment variable.
    """
    pkg_config = os.environ.get("PKG_CONFIG", "pkg-config")
    if shutil.which(pkg_config) is None:
        raise RuntimeError("Pkg-config is not installed. Adcc is not able to "
                           "find or autoinstall the libtensorlight library.")

    # Some default places to search for pkg-config files
    pkg_config_paths = [
        sysconfig.get_config_var("LIBDIR"), Path.home() / ".local" / "lib"
    ]
    for path in pkg_config_paths:
        if path is None:
            continue
        elif not isinstance(path, Path):
            path = Path(path)
        path = str(path / "pkgconfig")
        append_to_pkg_config_path(path)
    return pkg_config


def search_with_pkg_config(library: str, minversion=None,
                           define_prefix: bool = True):
    """
    Search the OS with pkg-config for a library and return the resulting
    cflags and libs stored inside the pc file. Also checks for a minimal
    version if `minversion` is not `None`.
    """
    pkg_config = get_pkg_config()

    cmd = [pkg_config, library]
    if define_prefix:
        cmd.append("--define-prefix")
    try:
        # first verify that the lib has the correct version
        if minversion:
            subprocess.check_call([*cmd, f"--atleast-version={minversion}"])
        # then get the include path for the headers for the compiler
        cflags = shlex.split(os.fsdecode(
            subprocess.check_output([*cmd, "--cflags"])
        ))
        # and the path to the library file and its name for the linker
        libs = shlex.split(os.fsdecode(
            subprocess.check_output([*cmd, "--libs"])
        ))
        return cflags, libs
    except (OSError, subprocess.CalledProcessError):
        return None, None


def extract_library_dirs(libs: list[str]) -> list[str]:
    """
    From the `libs` flags returned by `search_with_pkg_config` extract
    the existing library directories.
    """
    libdirs = []
    for flag in libs:
        if flag.startswith("-L") and Path(flag[2:]).is_dir():
            libdirs.append(flag[2:])
    return libdirs


def request_urllib(url: str, filename: str, token=None):
    """Download a file from the net"""
    import urllib.request

    # optionally add an auth token to the request
    request = urllib.request.Request(url)
    if token is not None:
        request.add_header("Authorization", token)

    try:
        resp = urllib.request.urlopen(request)
    except urllib.request.HTTPError as e:
        return e.code

    if 200 <= resp.status < 300:
        with open(filename, "wb") as fp:
            fp.write(resp.read())
    return resp.status


def assets_most_recent_release(project: str) -> list[str]:
    """
    Return the asset urls attached to the most recent release of a github project.
    """
    url = f"https://api.github.com/repos/{project}/releases"
    # look for the github token in the environment
    # and use it for the request to avoid the github API rate limits
    token = os.environ.get("GITHUB_TOKEN", None)
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = str(Path(tmpdir) / "releases.json")
        status = None
        for _ in range(10):
            status = request_urllib(url, filename, token=token)
            if 200 <= status < 300:
                break
            time.sleep(1)
        else:
            raise RuntimeError(f"Error downloading asset list from {url} "
                               f"... Response: {status}")

        with open(filename) as fp:
            ret = json.loads(fp.read())
        assets = [ret[i]["assets"] for i in range(len(ret))][0]
        return [asset["browser_download_url"] for asset in assets]


def url_most_recent_release(project: str) -> str:
    """
    Return the download url for the most recent release of a github project."""
    assets = assets_most_recent_release(project)
    if sys.platform == "linux":
        url = [asset for asset in assets if "-linux_x86_64" in asset]
    elif sys.platform == "darwin":
        # platform.machine() gives e.g., arm64
        # we want to match macosx_X_arm64, where X is the Version
        url = [
            asset for asset in assets
            if "-macosx_" in asset and f"_{platform.machine()}" in asset
        ]
    else:
        raise AssertionError("Can't download {project} from github"
                             " releases for unspported platform "
                             f"{sys.platform}.")
    if not url:
        raise RuntimeError(
            f"Could not find a version of project {project} to download. "
            "Check your platform is supported and if in doubt see the "
            "adcc installation instructions "
            "(https://adc-connect.org/latest/installation.html)."
        )
    assert len(url) == 1  # found more than 1 url for the current system??
    return url[0]


def install_libtensor(url: str, destination: str):
    """
    Download libtensor from `url` and install at `destination`, removing possibly
    existing files from a previous installation.
    """
    dest_folder = Path(destination)
    with tempfile.TemporaryDirectory() as tmpdir:
        print(f"Downloading libtensorlight from {url} to {destination} ...")
        tmpdir = Path(tmpdir).resolve()
        archive_file = tmpdir / Path(url).name

        # download the archive
        status_code = request_urllib(url, str(archive_file))  # add token?
        if status_code < 200 or status_code >= 300:
            raise RuntimeError(
                "Could not download libtensorlight. Check for a network issue "
                "and if in doubt see the adcc installation instructions "
                "(https://adc-connect.org/latest/installation.html)."
            )
        # remove old installation files
        file_globs = [
            "include/libtensorlight/**/*.h",
            "include/libtensor/**/*.h",
            "include/libutil/**/*.h",
            "lib/libtensorlight.so",
            "lib/libtensorlight.so.*",
            "lib/libtensorlight.dylib",
            "lib/libtensorlight.*.dylib",
            "lib/pkgconfig/libtensorlight.pc",
        ]
        for fglob in file_globs:
            for file in dest_folder.rglob(fglob):
                log.info(f"Removing old libtensor file: {str(file)}")
                assert file.is_file()
                file.unlink()
        # Change to installation directory and unpack the downloaded archive
        olddir = Path.cwd()
        dest_folder.mkdir(parents=True, exist_ok=True)
        os.chdir(destination)
        subprocess.run(["tar", "xf", archive_file], check=True)
        os.chdir(olddir)


def libadcc_sources(target: str) -> list[str]:
    # The paths have to be relative to the top level project directory
    # (the folder that contains setup.py - required by setuptools).
    # The compiler requires the source files relative to cwd.
    # Therefore, cwd == top_level_project_dir has to be enforced
    # for build and install to succeed.
    libadcc_src = Path("libadcc_src")
    test_src = libadcc_src / "tests"
    bindings_src = libadcc_src / "pyiface"
    assert libadcc_src.is_dir()
    assert test_src.is_dir()
    assert bindings_src.is_dir()
    # apparently older pybind versions expect strings
    sourcefiles = {str(p) for p in libadcc_src.rglob("*.cc")}
    unittests = [str(p) for p in test_src.rglob("*.cc")]
    exportfiles = [str(p) for p in bindings_src.rglob("*.cc")]
    if target == "extension":
        return sorted(sourcefiles.difference(unittests))
    elif target == "cpptest":
        return sorted(sourcefiles.difference(exportfiles))
    else:
        raise ValueError(f"Unknown target: {target}")


def update_flags_from_config(config_file: Path, *flags: dict):
    """
    Reads and executes the content of the config file and updates
    the prodided flags accordingly
    """
    log.info(f"Reading siteconfig file: {str(config_file)}")
    assert config_file.is_file()
    # merge the flags into a single dict: can't have the same keys!
    combined_flags = {}
    for flag_subset in flags:
        assert not flag_subset.keys() & combined_flags.keys()
        combined_flags.update(flag_subset)
    # read and execute the config and update the flags
    exec(open(config_file, "r").read(), combined_flags)
    # split the flags up in the original dicts only keeping
    # known keys that existed in the original dicts
    for key, val in combined_flags.items():
        for flag_subset in flags:
            if key in flag_subset:
                flag_subset[key] = val


@functools.lru_cache()
def libadcc_extension():
    # flags that are passed to the compiler
    build_flags: dict[str, list[str]] = {
        "libraries": [],
        "library_dirs": [],
        "include_dirs": [],
        "extra_link_args": [],
        "extra_compile_args": ["-Wall", "-Wextra", "-Werror", "-O3"],
        "runtime_library_dirs": [],
        "extra_objects": [],
        "define_macros": [],
    }
    # flags relevant for configuration of the extension (e.g. setting
    # the compiler flags and finding libtensor)
    config_flags = {
        "coverage": False,  # generate a test coverage report
        "search_system": True,  # search system for libtensorlight using pkg-config
        # install libtensor to folder if missing in the system
        "libtensor_autoinstall": None,
        # try to download libtensor from the url for the autoinstall
        "libtensor_url": None,
    }
    # only checked when we search the system or download the library!
    libtensorlight_min_version = "3.0.1"
    # The config files to check
    if "ADCC_CONFIG" in os.environ:
        if not Path(os.environ["ADCC_CONFIG"]).is_file():
            raise FileNotFoundError(os.environ["ADCC_CONFIG"])
        config_files: list[Path] = [Path(os.environ["ADCC_CONFIG"])]
    else:
        config_files: list[Path] = [
            Path("siteconfig.py"), Path.home() / ".adcc" / "siteconfig.py"
        ]

    # set specific compile args depending on the operating system
    if sys.platform == "darwin":
        build_flags["extra_compile_args"] += [
            "-Wno-unused-command-line-argument",
            "-Wno-undefined-var-template", "-Wno-bitwise-instead-of-logical"
        ]
        build_flags["extra_compile_args"].extend(["-arch", platform.machine()])
        build_flags["extra_link_args"].extend(["-arch", platform.machine()])
        # The build fails with clang 22.1.0 for macosx 11.0,
        # which is selected by default.
        # Specify 12.0 for now, but we might need to update in the
        # future once support for 12.0 is dropped, too.
        # Or remove the it again once the default is updated.
        if "MACOSX_DEPLOYMENT_TARGET" not in os.environ:
            os.environ["MACOSX_DEPLOYMENT_TARGET"] = "12.0"
    elif sys.platform.startswith("linux"):
        # otherwise fails with -O3 on gcc>=12
        build_flags["extra_compile_args"] += ["-Wno-array-bounds"]

    # folder to look for libtensor install: for linux and mac-os and if we are not
    # running in a conda environment (avoid creating a folder in home in conda)
    platform_autoinstall = (
        sys.platform.startswith("linux") or sys.platform.startswith("darwin")
    )
    if platform_autoinstall and not is_conda_build():
        # Keep '~' here and expand later to not change the behaviour!
        config_flags["libtensor_autoinstall"] = "~/.local"

    # User-provided config: modify build and config flags
    for siteconfig in config_files:
        if not siteconfig.is_file():
            continue
        # in-place update the flags from the config file
        update_flags_from_config(siteconfig, build_flags, config_flags)
        break  # only read a single config file!

    # Keep track whether libtensor has been found
    found_libtensor = "tensorlight" in build_flags["libraries"]

    if not found_libtensor:
        # Once we enter this if statement we require pkg-config to find
        # libtensorlight and generate the compiler and linker args!
        cflags, libs = None, None
        # first try to search the system with pkg-config
        if config_flags["search_system"]:
            log.info("Searching OS for libtensorlight using pkg-config")
            cflags, libs = search_with_pkg_config(
                "libtensorlight", libtensorlight_min_version
            )
        # if this was not successful we try to download libensorlight
        if (cflags is None or libs is None) and \
                config_flags["libtensor_autoinstall"]:
            # - get the url from where to download
            if config_flags["libtensor_url"]:
                url = config_flags["libtensor_url"]
            else:
                url = url_most_recent_release("adc-connect/libtensor")
            # download libtensor from the url and install it in the given folder
            install_dir = os.path.expanduser(config_flags["libtensor_autoinstall"])
            install_libtensor(url, install_dir)
            # add the installdir to the PKG_CONFIG_PATH so we can find the package
            append_to_pkg_config_path(
                str(Path(install_dir) / "lib" / "pkgconfig")
            )
            cflags, libs = search_with_pkg_config(
                "libtensorlight", libtensorlight_min_version
            )
            assert cflags is not None and libs is not None

        if cflags is not None and libs is not None:
            found_libtensor = True
            build_flags["extra_compile_args"].extend(cflags)
            build_flags["extra_link_args"].extend(libs)
            log.info(f"Using libtensorlight libraries: {libs}.")
            if sys.platform == "darwin":
                build_flags["extra_link_args"].append("-Wl,-rpath,@loader_path")
                for path in extract_library_dirs(libs):
                    build_flags["extra_link_args"].append(f"-Wl,-rpath,{path}")
            else:
                build_flags["runtime_library_dirs"].extend(
                    extract_library_dirs(libs)
                )

    if not found_libtensor:
        raise RuntimeError("Did not find the libtensorlight library.")

    ext = Pybind11Extension("libadcc", libadcc_sources("extension"),
                            language="c++", cxx_std=14, **build_flags)
    if config_flags["coverage"]:
        ext.extra_compile_args += [
            "--coverage", "-O0", "-g", "-fprofile-update=atomic"
        ]
        ext.extra_link_args += ["--coverage"]
    return ext


#
# Main setup code
#
def is_conda_build():
    return (
        os.environ.get("CONDA_BUILD", None) == "1"
        or os.environ.get("CONDA_EXE", None) is not None
    )


class build_ext(_Pybind11BuildExt):
    def run(self):
        super().run()
        # copy the libadcc stub file next to the executable to include it
        # in the bdist next to the executable
        pyi_src = Path(__file__).parent / "libadcc.pyi"
        if pyi_src.is_file():
            shutil.copy(str(pyi_src), str(Path(self.build_lib) / "libadcc.pyi"))


def read_readme():
    with open("README.md") as fp:
        return "".join([line for line in fp if not line.startswith("<img")])


if Path.cwd() != Path(__file__).parent or \
        not (Path("adcc") / "__init__.py").is_file():
    raise RuntimeError("Running setup.py is only supported "
                       "from top level of repository as './setup.py <command>'")

setup(
    # content of readme can't be modified from within pyproject.toml
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    #
    ext_modules=[libadcc_extension()],
    #
    cmdclass={
        "build_ext": build_ext, "build_docs": BuildDocs, "cpptest": CppTest
    },
)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np


def select_eigenpairs(eigenvalues, n_ep, which):
    """
    Return a numpy `bool` mask selecting the `n_ep` eigenpairs of the `which`
    criterion. It is assumed that the `eigenvalues` are sorted algebraically
    from the smallest to the largest.
    """
    mask = np.zeros(len(eigenvalues), dtype=bool)
    if which == "LA":    # Largest algebraic
        mask[-n_ep:] = True
    elif which == "SA":  # Smallest algebraic
        mask[:n_ep] = True
    elif which == "LM":  # Largest magnitude
        sorti = np.argsort(np.abs(eigenvalues))[-n_ep:]
        mask[sorti] = True
    elif which == "SM":  # Smallest magnitude
        sorti = np.argsort(np.abs(eigenvalues))[:n_ep]
        mask[sorti] = True
    else:
        raise ValueError("For now only the values 'LM', 'LA', 'SM' and 'SA' "
                         "are understood for 'which'.")
    return mask.nonzero()[0]



import os

# This file is loosely based upon the file
# cpp/ycm/.ycm_extra_conf.py from the youcompleteme daemon process
# available on github:
# https://github.com/Valloric/ycmd/blob/master/cpp/ycm/.ycm_extra_conf.py


# These are the compilation flags that will be used in case there's no
# compilation database set (by default, one is not set).
flags = [
    # Warnings: For a very detailed discussion about this
    # see the following stackexchange post:
    # https://programmers.stackexchange.com/questions/122608#124574
    '-Werror',
    '-Wall',
    '-Wextra',
    '-pedantic',
    "-Wnon-virtual-dtor",
    "-Woverloaded-virtual",
    "-Wshadow",
    "-Wold-style-cast",
    "-Wcast-align",
    "-Wconversion",
    "-Wuseless-cast",
    "-Wsign-conversion",
    "-Wmisleading-indentation",
    "-Wduplicated-cond",
    "-Wduplicated-branches",
    "-Wlogical-op",
    "-Wnull-dereference",
    "-Wdouble-promotion",
    "-Wformat=2",
    '-fexceptions',  # Generate unwind information
    '-DDEBUG',       # Compile as debug
    '-std=c++11',    # and c++ 11
    '-x', 'c++',     # Treat .h header files as c++
    # Include other libraries and show errors and
    # warnings within them
    # To suppress errors shown here, use "-isystem"
    # instead of "-I"
    '-isystem', '~/.local/include',
    '-isystem', '/usr/local/include',
    # Explicit clang includes:
    '-isystem', '/usr/include/c++/v1',
]


DIRECTORY_OF_THIS_SCRIPT = os.path.dirname(os.path.abspath(__file__))
SOURCE_EXTENSIONS = ['.cpp', '.cxx', '.cc', '.c', '.C']


def MakeRelativePathsInFlagsAbsolute(flags, working_directory):
    if not working_directory:
        return list(flags)
    new_flags = []
    make_next_absolute = False
    path_flags = ['-isystem', '-I', '-iquote', '--sysroot=']
    for flag in flags:
        new_flag = flag

        if make_next_absolute:
            make_next_absolute = False
            if not flag.startswith('/'):
                new_flag = os.path.join(working_directory, flag)

        for path_flag in path_flags:
            if flag == path_flag:
                make_next_absolute = True
                break

            if flag.startswith(path_flag):
                path = flag[len(path_flag):]
                new_flag = path_flag + os.path.join(working_directory, path)
                break

        if new_flag:
            new_flags.append(new_flag)
    return new_flags


def IsHeaderFile(filename):
    extension = os.path.splitext(filename)[1]
    return extension in ['.h', '.hxx', '.hpp', '.hh']


def FlagsForFile(filename, **kwargs):
    relative_to = DIRECTORY_OF_THIS_SCRIPT
    final_flags = MakeRelativePathsInFlagsAbsolute(flags, relative_to)

    return {
        'flags': final_flags,
        'do_cache': True
    }



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from os.path import basename

import h5py


def _emplace_ndarray(keyval, group, typ, **kwargs):
    dset = group.create_dataset(keyval[0], data=keyval[1], **kwargs)
    dset.attrs["type"] = "ndarray"


def _extract_ndarray(dataset):
    arr = np.empty(dataset.shape, dtype=dataset.dtype)
    dataset.read_direct(arr)

    if dataset.dtype == h5py.special_dtype(vlen=str):
        # HDF5 3.0.0 and up no longer extracts variable-string fields
        # as string but extracts them as raw bytes.
        # Here we decode the bytes explicitly.
        arr_flat = np.reshape(arr, -1)
        if len(arr_flat) > 0 and isinstance(arr_flat[0], bytes):
            arr_str = np.empty(dataset.shape, dtype='O')
            arr_str_flat = np.reshape(arr_str, -1)
            for i in range(len(arr_flat)):
                arr_str_flat[i] = arr_flat[i].decode()
            arr = arr_str
    return (basename(dataset.name), arr)


def _emplace_listlike(keyval, group, typ, **kwargs):
    dtype = None
    # Usually the heuristic for doing the conversion is pretty
    # good here, but there are some exceptions.
    if all(isinstance(v, str) for v in keyval[1]):
        dtype = h5py.special_dtype(vlen=str)

    ary = np.array(keyval[1], dtype=dtype)
    dset = group.create_dataset(keyval[0], data=ary, **kwargs)
    dset.attrs["type"] = "list"


def _extract_listlike(dataset):
    key, arr = _extract_ndarray(dataset)
    return (key, arr.tolist())


def _emplace_none(keyval, group, typ, **kwargs):
    dset = group.create_dataset(keyval[0], data=h5py.Empty("f"), **kwargs)
    dset.attrs["type"] = "none"


def _extract_none(dataset):
    return (basename(dataset.name), None)


# Type transformations for scalar types
# If type not found here, we have an error
# in the direction python -> hdf5, else we ignore it.
_scalar_transform = [
    (str,     h5py.special_dtype(vlen=str)),
    (bool,    np.dtype("b1")),
    (complex, np.dtype("c16")),
    (float,   np.dtype("f8")),
    (int,     np.dtype("int64")),
]


def _emplace_scalar(keyval, group, typ, compression=None, **kwargs):
    # Note: The compression key is present such that the compression
    # specification is silently dropped here and not passed onto
    # create_dataset
    dtype = None  # Indicate no target type found
    for t in _scalar_transform:
        if isinstance(keyval[1], t[0]):
            dtype = t[1]
            break
    if dtype is None:
        raise TypeError("Encountered unknown data type '"
                        + str(type(keyval[1])) + "'")

    dset = group.create_dataset(keyval[0], data=keyval[1],
                                dtype=dtype, **kwargs)
    dset.attrs["type"] = "scalar"


def _extract_scalar(dataset):
    dtype = None  # Target type to transform to
    for t in _scalar_transform:
        if dataset.dtype == t[1]:
            dtype = t[0]
            break

    if dataset.shape == ():  # i.e. HDF5 scalar
        ret = dataset[()]
    else:
        ret = dataset[0]

    if dtype == str and isinstance(ret, bytes):
        # HDF5 3.0.0 and up no longer extracts variable-string fields
        # as string but extracts them as raw bytes.
        ret = ret.decode()
    elif dtype is not None:
        ret = dtype(ret)
    return (basename(dataset.name), ret)


def _extract_dataset(dataset):
    """Select extractor based on the type attribute and use that
       to make the proper key-value pair out of the dataset
    """
    if "type" not in dataset.attrs:
        if dataset.shape == ():
            return _extract_scalar(dataset)  # Treat as scalar
        else:
            return _extract_ndarray(dataset)  # Treat as array
    else:
        # Use type attribute to distinguish what should happen
        tpe = dataset.attrs["type"]
        return {
            "scalar":   _extract_scalar,
            "none":     _extract_none,
            "ndarray":  _extract_ndarray,
            "list":     _extract_listlike,
            "tuple":    _extract_listlike,
        }[tpe](dataset)


def _emplace_key_value(kv, group, **kwargs):
    """
    Emplace a single key-value pair in the group.

    What precisely happends depends on the type of the value
    to emplace.
    """

    def _emplace_dict_inner(kv, group, typ, **kwargs):
        subgroup = group.create_group(kv[0])
        emplace_dict(kv[1], subgroup)

    emplace_map = [
        (np.ndarray,   _emplace_ndarray),
        (type(None),   _emplace_none),
        (list,         _emplace_listlike),
        (tuple,        _emplace_listlike),
        (dict,         _emplace_dict_inner),
    ]

    for (typ, emplace) in emplace_map:
        if isinstance(kv[1], typ):
            try:
                emplace(kv, group, typ, **kwargs)
            except TypeError as e:
                raise TypeError("Error with key '" + kv[0] + "': " + str(e))
            return

    # Fallback: Assume value is a simple scalar type
    try:
        _emplace_scalar(kv, group, typ, **kwargs)
    except TypeError as e:
        raise TypeError("Error with key '" + kv[0] + "': " + str(e))


#
# High-level routines
#
def emplace_dict(dictionary, group, **kwargs):
    """
    Emplace a python dictionary "d" into the HDF5 group "group"
    using the kwargs to create all neccessary datasets.
    """
    for kv in dictionary.items():
        _emplace_key_value(kv, group, **kwargs)


def extract_group(group):
    # Recursively extract all groups:
    ret = {basename(v.name): extract_group(v) for v in group.values()
           if isinstance(v, h5py.Group)}

    # Now deal with all datasets
    ret.update([_extract_dataset(v) for v in group.values()
                if isinstance(v, h5py.Dataset)])

    if not all(isinstance(v, (h5py.Dataset, h5py.Group)) or v is None
               for v in group.values()):
        raise ValueError("Encountered object in h5py which is neither "
                         "a Group nor a Dataset")
    return ret


def save(fname, dictionary):
    if not isinstance(dictionary, dict):
        raise TypeError("Second argument needs to be a dictionary")

    with h5py.File(fname, "w") as h5f:
        emplace_dict(dictionary, h5f, compression="gzip")


def load(fname):
    with h5py.File(fname, "r") as h5f:
        return extract_group(h5f)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

from import_data import import_data

# Gather preliminary data
data = import_data()

# Run an cvs-adc2 calculation:
singlets = adcc.cvs_adc2(data, core_orbitals=1, n_singlets=1, conv_tol=1e-8)
triplets = adcc.cvs_adc2(singlets.matrix, n_triplets=2, conv_tol=1e-8)
# Note: Above core_orbitals is not required again, since the precise CVS
#       splitting is already encoded in the matrix.

print(singlets.describe())
print(triplets.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc
import adcc.projection

from pyscf import gto, scf
from adcc.AdcMatrix import AdcMatrixProjected

# Aim of this script is to compute core excitations of water
# using a projection approach, i.e. where the non-CVS ADC matrix
# is projected into parts of the orbital subspaces to target core
# excitations specifically.

# Run SCF in pyscf
mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='cc-pvtz',
    unit="Bohr"
)
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-12
scfres.conv_tol_grad = 1e-9
scfres.kernel()
print(adcc.banner())

# Construct the standard ADC(2)-x matrix and project it, such that
# only cv, ccvv and ocvv excitations are allowed. Choose one orbital (oxygen 1s)
# to make up the core space.
matrix = adcc.AdcMatrix("adc2x", adcc.ReferenceState(scfres))
pmatrix = AdcMatrixProjected(matrix, ["cv", "ccvv", "ocvv"], core_orbitals=1)

# Run the ADC calculation. Note that doubles guesses are not yet supported in this
# setup and like lead to numerical issues (zero excitations) if selected. So we
# explicitly disable them here.
state_proj = adcc.run_adc(pmatrix, n_singlets=7, conv_tol=1e-8, n_guesses_doubles=0)
print(state_proj.describe())

# For comparison we also run a standard CVS-ADC(2)-x calculation:
state_cvs = adcc.cvs_adc2x(scfres, n_singlets=7, conv_tol=1e-8, core_orbitals=1)
print(state_cvs.describe())

# Note that this calculation could also be used as an initial guess:
guesses = adcc.projection.transfer_cvs_to_full(state_cvs, pmatrix)
state_proj2 = adcc.run_adc(pmatrix, n_singlets=7, conv_tol=1e-8, guesses=guesses)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2022 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import libadcc
import warnings
import itertools

import numpy as np

from .guess import guess_kwargs_kind, guess_symmetries
from .Tensor import Tensor
from .MoSpaces import expand_spaceargs
from .Symmetry import Symmetry
from .AdcMatrix import AdcMatrixlike
from .AmplitudeVector import AmplitudeVector

__all__ = ["SubspacePartitioning", "Projector"]


class SubspacePartitioning:
    def __init__(self, mospaces, core_orbitals=None, outer_virtuals=None,
                 partitions=None, aliases=None):
        """
        Partition the "o1" and "v1" occupied and virtual orbital subspace into two
        (or more) parts. In combination with the :py:`adcc.Projector`
        class this allows to disable some excitations in an amplitude vector.

        Parameters
        ----------
        mospaces : MoSpaces
            MoSpaces to base partitioning on.
        core_orbitals : int or list or tuple, optional
            The orbitals to be put into the ``c`` core space.
            For details on the option how to select these spaces, see the
            documentation in :py:`adcc.ReferenceState.__init__` (``outer_virtuals``
            follows the same rules as ``frozen_virtuals``).
        outer_virtuals : int or list or tuple, optional
            The virtuals to be put into the ``w`` outer-virtual space.
        partitions : dict, optional
            Partitioning of the spaces: Mapping from the partitioned spaces to
            the indices of the orbitals which belong to this partition. The
            indices are given relative to the corresponding full orbital subspace
            (advanced argument).
        aliases : dict, optional
            Aliase mapping to give partitions a more user-friendly identifier
            (advanced argument).
        """
        if mospaces.has_core_occupied_space or "v2" in mospaces.subspaces_virtual:
            raise ValueError("Cannot use MoSpaces with core-valence separation "
                             "or frozen virtuals.")
        assert mospaces.subspaces_occupied == ["o1"]

        if not partitions or not aliases:
            if core_orbitals is None and outer_virtuals is None:
                raise ValueError("One of core_orbitals or outer_virtuals should "
                                 "be passed.")

            # core / outer virtual part of the "normal" occupied / virtual subspace
            noa = mospaces.n_orbs_alpha("o1")
            nob = mospaces.n_orbs_beta("o1")
            nva = mospaces.n_orbs_alpha("v1")
            nvb = mospaces.n_orbs_beta("v1")
            core = expand_spaceargs((noa, nob), core_orbitals=core_orbitals)
            core = core["core_orbitals"]
            fv = expand_spaceargs((nva, nvb), frozen_virtual=outer_virtuals)
            fv = fv["frozen_virtual"]

            # The active core and virtual parts:
            occ = [i for i in range(mospaces.n_orbs("o1")) if i not in core]
            av = [i for i in range(mospaces.n_orbs("v1")) if i not in fv]
            assert occ and av

            # Setup the space partitionings and aliases
            partitions = {"o1.1": occ, "v1.1": av, }
            aliases = {"o": "o1.1", "v": "v1.1", }
            if core:
                partitions["o1.2"] = core
                aliases["c"] = "o1.2"
            if fv:
                partitions["v1.2"] = fv
                aliases["w"] = "v1.2"
            self.__init__(mospaces, partitions=partitions, aliases=aliases)
        else:
            assert core_orbitals is None
            assert outer_virtuals is None

            self.partitions = partitions
            self.aliases = aliases
            self.mospaces = mospaces

            # Determine whether the partitioning keeps spin symmetry.
            self.keeps_spin_symmetry = mospaces.restricted
            for (label, indices) in partitions.items():
                if not self.keeps_spin_symmetry:
                    break
                nalpha = mospaces.n_orbs_alpha(label[:2])
                alpha_part = [i for i in indices if i < nalpha]
                beta_part = [i - nalpha for i in indices if i >= nalpha]
                self.keeps_spin_symmetry = alpha_part == beta_part

            if mospaces.restricted and not self.keeps_spin_symmetry:
                warnings.warn(
                    "For restricted references the case of a space "
                    "partitioning, which breaks spin (i.e. where excitations "
                    "differing only in spin are placed in different "
                    "partitions) has not been tested. You are on your own."
                )

    def list_space_partitions(self, space):
        """
        Return the partitions into which the passed space is split.
        Resolves aliases, i.e. does not return not the list of strings
        such as ['o1.1', 'o1.2], but the aliased equivalent ["o", "c"].
        """
        partitions = [s for s in self.partitions if s.startswith(space + ".")]
        partitions = [key for (key, value) in self.aliases.items()
                      if value in partitions]
        return sorted(partitions)

    def get_partition(self, label):
        """Return the index list corresponding to a partition label"""
        return self.partitions[self.aliases[label]]


class Projector:
    def __init__(self, subspaces, partitioning, blocks_to_keep):
        """
        Initialise a projector, which upon multiplication with a tensor sets
        a number of partitioning blocks of this tensor to zero. These are defined
        by the passed `partitioning` and the list of `blocks_to_keep`.

        Parameters
        ----------
        subspaces : list
            List of subspaces to act upon (e.g. ``["o1", "v1", "v1"]``)
        partitioning : SubspacePartitioning
            Partitioning of the ``o1`` occupied and ``v1`` virtual space.
        blocks_to_keep : list
            Blocks which are kept as non-zero (e.g. ``["cv", "ccvv", "ocvv"]``).
        """
        if len(subspaces) == 3 or len(subspaces) > 4 \
                or (len(subspaces) > 2 and subspaces[1] == subspaces[2]):
            raise NotImplementedError(
                f"Projector not implemented for subspaces {subspaces}"
            )

        def normalise_block(labels):
            labels = list(labels)
            if len(subspaces) <= 1:
                return labels
            if subspaces[0] == subspaces[1] and labels[0] > labels[1]:
                labels[0], labels[1] = labels[1], labels[0]
            if len(subspaces) <= 2:
                return labels
            if len(subspaces) == 3 or subspaces[1] == subspaces[2] \
                    or len(subspaces) > 4:
                raise NotImplementedError("not implemented")
            if subspaces[2] == subspaces[3] and labels[2] > labels[3]:
                labels[2], labels[3] = labels[3], labels[2]
            return labels

        blocks_to_keep = ["".join(normalise_block(block))
                          for block in blocks_to_keep]
        partitions = [partitioning.list_space_partitions(space)
                      for space in subspaces]
        all_partition_blocks = ["".join(normalise_block(block))
                                for block in itertools.product(*partitions)]
        for block in blocks_to_keep:
            if block not in all_partition_blocks:
                raise ValueError(f"Partition block {block} not known.")

        sym = Symmetry(partitioning.mospaces, "".join(subspaces))
        sym.irreps_allowed = ["A"]
        if partitioning.keeps_spin_symmetry:
            if len(subspaces) == 1:
                sym.spin_block_maps = [("a", "b", 1)]
            elif len(subspaces) == 2:
                sym.spin_block_maps = [("aa", "bb", 1), ("ab", "ba", 1)]
            elif len(subspaces) == 4:
                sym.spin_block_maps = [
                    ("aaaa", "bbbb", 1), ("aaab", "bbba", 1),
                    ("aaba", "bbab", 1), ("abaa", "babb", 1),
                    ("baaa", "abbb", 1), ("aabb", "bbaa", 1),
                    ("abba", "baab", 1), ("abab", "baba", 1)
                ]
            else:
                raise NotImplementedError("not implemented")
        if len(subspaces) == 4:
            assert subspaces[0] == subspaces[1]
            assert subspaces[2] == subspaces[3]
            sym.permutations = ["ijkl", "jikl", "ijlk"]

        # Allocate zero tensor to hold projection kernel:
        kernel = Tensor(sym)

        # Generic implementation not using symmetry
        for block in blocks_to_keep:
            ranges = [partitioning.get_partition(b) for b in block]
            for index in itertools.product(*ranges):
                kernel[index] = 1.0

        self.subspaces = subspaces
        self.partitioning = partitioning
        self.blocks_to_keep = blocks_to_keep
        self.kernel = kernel

    def matvec(self, v):
        return self.kernel * v

    def __matmul__(self, other):
        if isinstance(other, libadcc.Tensor):
            return self.matvec(other)
        if isinstance(other, list):
            if all(isinstance(elem, libadcc.Tensor) for elem in other):
                return [self.matvec(ov) for ov in other]
        return NotImplemented


def transfer_amplitude_cvs_to_full(mospaces_cvs, cvs, symmetry_full, tol=1e-12):
    """
    Take a CVS vector `cvs` taken from the MoSpace `mospaces_cvs` and copy
    its data into a "full" ADC amplitude constructed according to the symmetry
    `symmetry_full`. Notice that this will fail if the symmetry of `cvs` is
    not compatible (e.g. differs in targeted spin manifold, restricted versus
    unrestricted etc.).
    """
    if cvs.keys() != symmetry_full.keys():
        raise ValueError("Blocks present in CVS and full vector should agree.")
    if not mospaces_cvs.has_core_occupied_space:
        raise ValueError("Should have CVS enabled in mospaces_cvs")
    if not isinstance(cvs, AmplitudeVector):
        raise TypeError("Expected cvs to be an AmplitudeVector.")
    mospaces_full = symmetry_full["ph"].mospaces
    n_c_a = mospaces_cvs.n_orbs_alpha("o2")
    n_o_a = mospaces_cvs.n_orbs_alpha("o1")
    n_of_a = mospaces_full.n_orbs_alpha("o1")

    cvs_contiguous_in_full = (
        mospaces_cvs.core_orbitals[:n_c_a]
        == mospaces_full.occupied_orbitals[:n_c_a]
        and mospaces_cvs.core_orbitals[n_c_a:]
        == mospaces_full.occupied_orbitals[n_of_a:n_of_a + n_c_a]
    )
    if not cvs_contiguous_in_full:
        raise NotImplementedError(
            "For transfer CVS to full only the case where the CVS space is "
            "contiguous in the full space has been implemented, i.e. when the "
            "`core_orbitals` are chosen as an integer in the run_adc routines."
        )
    full = AmplitudeVector(**{block: Tensor(sym)
                              for block, sym in symmetry_full.items()})

    # Singles block
    assert cvs.ph.subspaces == ["o2", "v1"]
    assert full.ph.subspaces == ["o1", "v1"]
    assert cvs.ph.shape == (mospaces_cvs.n_orbs("o2"), mospaces_cvs.n_orbs("v1"))

    cvs_ph = cvs.ph.to_ndarray()
    full_ph = np.zeros(full.ph.shape)
    full_ph[0:n_c_a, :] = cvs_ph[:n_c_a, :]
    full_ph[n_of_a:n_of_a + n_c_a, :] = cvs_ph[n_c_a:, :]
    full.ph.set_from_ndarray(full_ph, tol)

    # Doubles block
    if "pphh" in cvs:
        assert cvs.pphh.subspaces == ["o1", "o2", "v1", "v1"]
        assert full.pphh.subspaces == ["o1", "o1", "v1", "v1"]

        # Note 1/sqrt(2) factor is there because the ocvv ocurrs twice
        # in the full amplitude, but only once in the CVS amplitude.
        cvs_pphh = cvs.pphh.to_ndarray() / np.sqrt(2)
        cvs_pphh_aa = cvs_pphh[:n_o_a, :n_c_a, :, :]
        cvs_pphh_ab = cvs_pphh[:n_o_a, n_c_a:, :, :]
        cvs_pphh_ba = cvs_pphh[n_o_a:, :n_c_a, :, :]
        cvs_pphh_bb = cvs_pphh[n_o_a:, n_c_a:, :, :]

        full_pphh = np.zeros(full.pphh.shape)
        full_pphh[n_c_a:n_of_a, 0:n_c_a, :, :] = cvs_pphh_aa
        full_pphh[0:n_c_a, n_c_a:n_of_a, :, :] = (
            -cvs_pphh_aa.transpose((1, 0, 2, 3))
        )
        full_pphh[n_c_a:n_of_a, n_of_a:n_of_a + n_c_a, :, :] = cvs_pphh_ab
        full_pphh[n_of_a:n_of_a + n_c_a, n_c_a:n_of_a, :, :] = (
            -cvs_pphh_ab.transpose((1, 0, 2, 3))
        )
        full_pphh[n_of_a + n_c_a:, 0:n_c_a, :, :] = cvs_pphh_ba
        full_pphh[0:n_c_a, n_of_a + n_c_a:, :, :] = (
            -cvs_pphh_ba.transpose((1, 0, 2, 3))
        )
        full_pphh[n_of_a + n_c_a:, n_of_a:n_of_a + n_c_a, :, :] = cvs_pphh_bb
        full_pphh[n_of_a:n_of_a + n_c_a, n_of_a + n_c_a:, :, :] = (
            -cvs_pphh_bb.transpose((1, 0, 2, 3))
        )
        full.pphh.set_from_ndarray(full_pphh, tol)
    return full


def transfer_cvs_to_full(state_matrix_cvs, matrix_full, vector=None, kind=None,
                         spin_change=0, spin_block_symmetrisation="none"):
    """
    Transfer `vector` (or a list of vectors) from the CVS space to the full space
    defined by the ADC matrix `matrix_full`. `state_matrix_cvs` is either the
    CVS matrix or the state containing the solved CVS excitations. To properly
    set up the symmetry of the returned vectors, set the `kind`, `spin_change`
    and `spin_block_symmetrisation` parameters. If the first argument is an
    `ExcitedStates` object, the symmetry is automatically detected.
    """
    if kind is None and hasattr(state_matrix_cvs, "kind"):
        kind = state_matrix_cvs.kind

    if spin_change is None and spin_block_symmetrisation is None:
        if kind is None:
            raise ValueError("kind needs to be given if first argument is not an "
                             "ExcitedStates object and spin symmetry setup is not "
                             "explicitly given.")
        return transfer_cvs_to_full(state_matrix_cvs, matrix_full, vector, kind,
                                    **guess_kwargs_kind(kind))

    if vector is None:
        if hasattr(state_matrix_cvs, "excitation_vector"):
            vector = state_matrix_cvs.excitation_vector
        else:
            raise ValueError("vector needs to be given if first argument is not an "
                             "ExcitedStates object.")
    if isinstance(vector, list):
        return [transfer_cvs_to_full(state_matrix_cvs, matrix_full, v, kind,
                                     **guess_kwargs_kind(kind)) for v in vector]

    if isinstance(state_matrix_cvs, AdcMatrixlike):
        mospaces_cvs = state_matrix_cvs.mospaces
    else:
        mospaces_cvs = state_matrix_cvs.matrix.mospaces

    sym = guess_symmetries(matrix_full, spin_change=spin_change,
                           spin_block_symmetrisation=spin_block_symmetrisation)
    return transfer_amplitude_cvs_to_full(mospaces_cvs, vector, sym)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2026 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from .NParticleDensity import NParticleDensity
from .NParticleOperator import OperatorSymmetry


class TwoParticleDensity(NParticleDensity):
    def __init__(self, spaces, symmetry=OperatorSymmetry.HERMITIAN):
        """
        Construct a TwoParticleDensity object. All blocks are initialised
        as zero blocks.

        Parameters
        ----------
        spaces : adcc.MoSpaces or adcc.ReferenceState or adcc.LazyMp
            MoSpaces object

        symmetry : OperatorSymmetry, optional
        """
        super().__init__(spaces, n_particle_op=2, symmetry=symmetry)

    def _construct_empty(self):
        """
        Create an empty instance of a TwoParticleDensity
        """
        return TwoParticleDensity(
            self.mospaces,
            symmetry=self.symmetry,
        )

    def _transform_to_ao(self, refstate_or_coefficients):
        """
        MO -> AO transformation of antisymmetrized densities/integrals is
        not implemented, because information about spin blocks is irreversibly lost
        after antisymmetrization. Example: Product Trace of a TwoParticleDensity
        with an ERI:
        aaaa block: + Gamma^pq_rs <pq||rs>
        bbbb block: + Gamma^pq_rs <pq||rs>
        abab block: + Gamma^pq_rs <pq|rs>
        abba block: - Gamma^pq_rs <pq|sr>
        """
        raise NotImplementedError(
            "MO -> AO transformation of antisymmetrized densities/integrals is "
            "not implemented, because information about spin blocks is "
            "irreversibly lost after antisymmetrization."
        )



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from itertools import product
from collections import namedtuple


def range_in(inner, full):
    if inner.start is None:
        inner = slice(0, inner.stop, 1)
    if full.start is None:
        full = slice(0, full.stop, 1)
    return all(r in range(full.start, full.stop)
               for r in range(inner.start, inner.stop))


# Helper namedtuple for slices of spin blocks
SpinBlockSlice = namedtuple('SpinBlockSlice',
                            ['block', 'spin', 'fromslice', 'toslice'])


class EriBuilder:
    """
    Parent class for building ERIs with different backends

    Implementation of the following functions in a derived class
    is necessary:
        - ``compute_mo_eri``: compute a block of integrals (Chemists' notation)
          Gets passed the block as a string like 'OOVV' and the spin block as
          as string like 'abab'.
    """
    def __init__(self, n_orbs, n_orbs_alpha, n_alpha, n_beta, restricted):
        self.n_orbs = n_orbs
        self.n_orbs_alpha = n_orbs_alpha
        self.n_alpha = n_alpha
        self.n_beta = n_beta
        self.eri_cache = {}
        self.restricted = restricted
        self.block2slice = {
            "oa": slice(0, self.n_alpha, 1),
            "va": slice(self.n_alpha, self.n_orbs_alpha, 1),
            "ob": slice(self.n_orbs_alpha, self.n_orbs_alpha + self.n_beta, 1),
            "vb": slice(self.n_orbs_alpha + self.n_beta, self.n_orbs, 1),
        }

    def compute_mo_eri(self, blocks, spins):
        """
        Compute block of the ERI tensor in chemists' indexing
        """
        raise NotImplementedError("Implement compute_mo_eri")

    def split_4d_slice(self, slices):
        """
        Split tuple of four slices into the block spin slices
        and their mapping to where elements are to be placed
        """
        return [SpinBlockSlice(tpl[0][0] + tpl[1][0] + tpl[2][0] + tpl[3][0],
                               tpl[0][1] + tpl[1][1] + tpl[2][1] + tpl[3][1],
                               (tpl[0][2], tpl[1][2], tpl[2][2], tpl[3][2]),
                               (tpl[0][3], tpl[1][3], tpl[2][3], tpl[3][3]))
                for tpl in product(*(self.split_1d_slice(sl) for sl in slices))]

    def split_1d_slice(self, sl):
        """
        Split slice into block-slices or multiple block-slices
        """
        if sl.start is None:
            sl = slice(0, sl.stop, 1)
        if sl.step is None:
            sl = slice(sl.start, sl.stop, 1)

        ret = []
        for (block, bslice) in self.block2slice.items():
            fromslice = toslice = None
            if range_in(sl, bslice):
                fromslice = (sl.start - bslice.start, sl.stop - bslice.start)
                toslice = (0, sl.stop - sl.start)
            elif range_in(bslice, sl):
                fromslice = (0, bslice.stop - bslice.start)
                toslice = (bslice.start - sl.start, bslice.stop - sl.start)
            elif sl.start in range(bslice.start, bslice.stop, 1):
                # Because the previous if failed, it cannot be the full range
                fromslice = (sl.start - bslice.start, bslice.stop - bslice.start)
                toslice = (0, bslice.stop - sl.start)
            elif (sl.stop - 1) in range(bslice.start, bslice.stop, 1):
                # Because the previous ifs failed, it cannot be the full range
                fromslice = (0, sl.stop - bslice.start)
                toslice = (bslice.start - sl.start, sl.stop - sl.start)
            if fromslice is None or toslice is None:
                continue   # Not found
            ret.append(SpinBlockSlice(block[0].upper(), block[1],
                                      slice(*fromslice), slice(*toslice)))
        assert len(ret) > 0
        return ret

    def fill_slice_symm(self, slices, out):
        for sbslices in self.split_4d_slice(slices):
            blocks, spins, fromslices, toslices = sbslices
            if spins not in ["aaaa", "aabb", "bbaa", "bbbb"]:
                out[toslices] = 0  # Zero by symmetry
                continue
            if self.restricted:
                # For restricted spins in chem eri do not matter
                spins = "aaaa"

            cache_key = "".join(blocks) + "".join(spins)
            if cache_key in self.eri_cache:
                eri = self.eri_cache[cache_key]
            else:
                eri = self.compute_mo_eri(blocks, spins)
                self.eri_cache[cache_key] = eri

            out[toslices] = eri[fromslices]

    def flush_cache(self):
        self.eri_cache = {}



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import os
import adcc
import numpy as np

from pyscf import gto, scf
from matplotlib import pyplot as plt


def run_spin_flip(distance):
    # Run SCF in pyscf
    mol = gto.M(
        atom='H 0 0 0;'
             'F 0 0 {}'.format(distance),
        basis='6-31G',
        unit="Bohr",
        spin=2  # =2S, ergo triplet
    )
    scfres = scf.UHF(mol)
    scfres.conv_tol = 1e-12
    scfres.conv_tol_grad = 1e-8
    scfres.kernel()

    # Run ADC and compute total energy
    states = adcc.adc2(scfres, n_spin_flip=1)

    ene = scfres.energy_tot() + states.ground_state.energy_correction(2)
    return ene + states.excitation_energy[0]


def run_progression(outfile="631g_adc2_dissociation.nptxt"):
    if os.path.isfile(outfile):
        return np.loadtxt(outfile)

    dists = np.linspace(1.0, 5.0, 30)
    enes = [run_spin_flip(d) for d in dists]

    result = np.vstack((dists, enes)).T
    np.savetxt(outfile, result)
    return result


def main():
    result = run_progression()
    plt.plot(result[:, 0], result[:, 1])
    plt.show()


if __name__ == "__main__":
    main()
